"""
This module contains a very trimmed-down version of fbs's built-in commands.
"""
from fbs import path, SETTINGS
from fbs.builtin_commands._util import prompt_for_value, \
    require_existing_project, require_frozen_app
from fbs.cmdline import command
from fbs.resources import copy_with_filtering
from fbs_runtime import FbsError
from fbs_runtime.platform import is_windows, is_mac, is_linux, is_arch_linux, \
    is_ubuntu, is_fedora
from fbs_tutorial_shim import unpack, UnknownHash
from os import listdir, remove, unlink, mkdir
from os.path import join, isfile, isdir, islink, dirname, exists
from shutil import rmtree

import logging
import os
import subprocess
import sys

_LOG = logging.getLogger(__name__)

_FIXED_APP_NAME = 'MyFbsApp'
_FIXED_AUTHOR = 'Author'
_FIXED_MAC_BUNDLE_IDENTIFIER = 'com.author.myfbsapp'
_FIXED_PYTHON_BINDINGS = 'PyQt5'

@command
def startproject():
    """
    Start a new project in the current directory
    """
    if exists('src'):
        raise FbsError('The src/ directory already exists. Aborting.')
    mkdir('src')
    template_dir = join(dirname(__file__), 'project_template')
    template_path = lambda relpath: join(template_dir, *relpath.split('/'))
    # fbs prompts the user for several values such as the app name.
    # We use hard-coded values to be able to use pre-compiled binaries.
    copy_with_filtering(
        template_dir, '.', {
            'app_name': _FIXED_APP_NAME,
            'author': _FIXED_AUTHOR,
            'mac_bundle_identifier': _FIXED_MAC_BUNDLE_IDENTIFIER,
            'python_bindings': _FIXED_PYTHON_BINDINGS
        },
        files_to_filter=[
            template_path('src/build/settings/base.json'),
            template_path('src/build/settings/mac.json'),
            template_path('src/main/python/main.py')
        ]
    )
    print('')
    _LOG.info(
        "Created the src/ directory. You can now do:\n\n"
        "    fbs run"
    )

@command
def run():
    """
    Run your app from source
    """
    require_existing_project()
    env = dict(os.environ)
    pythonpath = path('src/main/python')
    old_pythonpath = env.get('PYTHONPATH', '')
    if old_pythonpath:
        pythonpath += os.pathsep + old_pythonpath
    env['PYTHONPATH'] = pythonpath
    subprocess.run([sys.executable, path(SETTINGS['main_module'])], env=env)

@command
def freeze():
    """
    Compile your code to a standalone executable
    """
    require_existing_project()
    app_name = SETTINGS['app_name']
    main_py = path('src/main/python/main.py')
    if is_mac():
        import fbs_tutorial_shim_mac as os_shim_module
        executable = 'target/%s.app/Contents/MacOS/%s' % (app_name, app_name)
    elif is_windows():
        import fbs_tutorial_shim_windows as os_shim_module
        executable = join('target', app_name, app_name) + '.exe'
    else:
        # It would be nice to support Linux as well. But our current approach of
        # hosting OS-specific frozen binaries on PyPI does not allow it. Here's
        # why: We obtain OS-specific binaries by declaring them as dependencies
        # in setup.py. For example:
        #     fbs-tutorial-shim-windows; sys_platform=='windows'
        # For Linux, we need different binaries depending on the distribution.
        # However, sys_platform only gives us "linux". This means that we would
        # need to package the binaries for all Linux distributions in one PyPI
        # package. But here comes the problem: PyPI has a maximum file size of
        # 60 MB. This is not enough to contain the files for multiple
        # distributions.
        raise FbsError(
            "This copy of fbs does not support Linux, sorry.\n"
            "Please use Python 3.5/3.6 and `pip install fbs PyQt5==5.9.2`.\n"
            "Or obtain fbs Pro from https://build-system.fman.io/pro."
        )
    data_dir = join(dirname(os_shim_module.__file__), 'data')
    try:
        unpack(main_py, data_dir, path('${freeze_dir}'))
    except UnknownHash:
        raise FbsError(
            "This copy of fbs only supports main.py as in the tutorial.\n"
            "Please use Python 3.5/3.6 and `pip install fbs PyQt5==5.9.2`.\n"
            "Or obtain fbs Pro from https://build-system.fman.io/pro."
        )
    _LOG.info(
        "Done. You can now run `%s`. If that doesn't work, see "
        "https://build-system.fman.io/troubleshooting.", executable
    )

@command
def installer():
    """
    Create an installer for your app
    """
    require_frozen_app()
    linux_distribution_not_supported_msg = \
        "Your Linux distribution is not supported, sorry. " \
        "You can run `fbs buildvm` followed by `fbs runvm` to start a Docker " \
        "VM of a supported distribution."
    try:
        installer_fname = SETTINGS['installer']
    except KeyError:
        if is_linux():
            raise FbsError(linux_distribution_not_supported_msg)
        raise
    out_file = join('target', installer_fname)
    msg_parts = ['Created %s.' % out_file]
    if is_windows():
        from fbs.installer.windows import create_installer_windows
        create_installer_windows()
    elif is_mac():
        from fbs.installer.mac import create_installer_mac
        create_installer_mac()
    elif is_linux():
        app_name = SETTINGS['app_name']
        if is_ubuntu():
            from fbs.installer.ubuntu import create_installer_ubuntu
            create_installer_ubuntu()
            install_cmd = 'sudo dpkg -i ' + out_file
            remove_cmd = 'sudo dpkg --purge ' + app_name
        elif is_arch_linux():
            from fbs.installer.arch import create_installer_arch
            create_installer_arch()
            install_cmd = 'sudo pacman -U ' + out_file
            remove_cmd = 'sudo pacman -R ' + app_name
        elif is_fedora():
            from fbs.installer.fedora import create_installer_fedora
            create_installer_fedora()
            install_cmd = 'sudo dnf install ' + out_file
            remove_cmd = 'sudo dnf remove ' + app_name
        else:
            raise FbsError(linux_distribution_not_supported_msg)
        msg_parts.append(
            'You can for instance install it via the following command:\n'
            '    %s\n'
            'This places it in /opt/%s. To uninstall it again, you can use:\n'
            '    %s'
            % (install_cmd, app_name, remove_cmd)
        )
    else:
        raise FbsError('Unsupported OS')
    _LOG.info(' '.join(msg_parts))

@command
def clean():
    """
    Remove previous build outputs
    """
    try:
        rmtree(path('target'))
    except FileNotFoundError:
        return
    except OSError:
        # In a docker container, target/ may be mounted so we can't delete it.
        # Delete its contents instead:
        for f in listdir(path('target')):
            fpath = join(path('target'), f)
            if isdir(fpath):
                rmtree(fpath, ignore_errors=True)
            elif isfile(fpath):
                remove(fpath)
            elif islink(fpath):
                unlink(fpath)

_REQUIRES_FULL_FBS_COPY = \
    "This copy of fbs only supports the tutorial commands, sorry.\n" \
    "Please use Python 3.5/3.6 and `pip install fbs PyQt5==5.9.2`.\n" \
    "Or obtain fbs Pro from https://build-system.fman.io/pro."

@command
def sign():
    """
    Sign your app, so the user's OS trusts it
    """
    raise FbsError(_REQUIRES_FULL_FBS_COPY)

@command
def sign_installer():
    """
    Sign installer, so the user's OS trusts it
    """
    raise FbsError(_REQUIRES_FULL_FBS_COPY)

@command
def repo():
    """
    Generate files for automatic updates
    """
    raise FbsError(_REQUIRES_FULL_FBS_COPY)

@command
def upload():
    """
    Upload installer and repository to fbs.sh
    """
    raise FbsError(_REQUIRES_FULL_FBS_COPY)

@command
def release(version=None):
    """
    Bump version and run clean,freeze,...,upload
    """
    raise FbsError(_REQUIRES_FULL_FBS_COPY)

@command
def test():
    """
    Execute your automated tests
    """
    raise FbsError(_REQUIRES_FULL_FBS_COPY)

@command
def buildvm():
    """
    Build a Linux VM. Eg.: buildvm ubuntu
    """
    raise FbsError(_REQUIRES_FULL_FBS_COPY)

@command
def runvm():
    """
    Run a Linux VM. Eg.: runvm ubuntu
    """
    raise FbsError(_REQUIRES_FULL_FBS_COPY)

@command
def gengpgkey():
    """
    Generate a GPG key for Linux code signing
    """
    raise FbsError(_REQUIRES_FULL_FBS_COPY)
@command
def register():
    """
    Create an account for uploading your files
    """
    raise FbsError(_REQUIRES_FULL_FBS_COPY)
@command
def login():
    """
    Save your account details to secret.json
    """
    raise FbsError(_REQUIRES_FULL_FBS_COPY)
@command
def init_licensing():
    """
    Generate public/private keys for licensing
    """
    raise FbsError(_REQUIRES_FULL_FBS_COPY)


from fbs import _state
from fbs._state import LOADED_PROFILES
from fbs_runtime import FbsError, _source
from fbs_runtime._fbs import get_core_settings, get_default_profiles
from fbs_runtime._settings import load_settings, expand_placeholders
from fbs_runtime._source import get_settings_paths
from os.path import abspath

import sys

"""
fbs populates SETTINGS with the current build settings. A typical example is
SETTINGS['app_name'], which you define in src/build/settings/base.json.
"""
SETTINGS = _state.SETTINGS

def init(project_dir):
    """
    Call this if you are invoking neither `fbs` on the command line nor
    fbs.cmdline.main() from Python.
    """
    SETTINGS.update(get_core_settings(abspath(project_dir)))
    for profile in get_default_profiles():
        activate_profile(profile)

def activate_profile(profile_name):
    """
    By default, fbs only loads some settings. For instance,
    src/build/settings/base.json and .../`os`.json where `os` is one of "mac",
    "linux" or "windows". This function lets you load other settings on the fly.
    A common example would be during a release, where release.json contains the
    production server URL instead of a staging server.
    """
    LOADED_PROFILES.append(profile_name)
    project_dir = SETTINGS['project_dir']
    json_paths = get_settings_paths(project_dir, LOADED_PROFILES)
    core_settings = get_core_settings(project_dir)
    SETTINGS.update(load_settings(json_paths, core_settings))

def path(path_str):
    """
    Return the absolute path of the given file in the project directory. For
    instance: path('src/main/python'). The `path_str` argument should always use
    forward slashes `/`, even on Windows. You can use placeholders to refer to
    settings. For example: path('${freeze_dir}/foo').
    """
    path_str = expand_placeholders(path_str, SETTINGS)
    try:
        project_dir = SETTINGS['project_dir']
    except KeyError:
        error_message = "Cannot call path(...) until fbs.init(...) has been " \
                        "called."
        raise FbsError(error_message) from None
    return _source.path(project_dir, path_str)


from argparse import ArgumentParser
from fbs._state import COMMANDS
from fbs_runtime import FbsError
from inspect import getfullargspec
from os import getcwd
from os.path import basename, splitext

import fbs
import logging
import sys

_LOG = logging.getLogger(__name__)

def main(project_dir=None):
    """
    This function is executed when you run `fbs ...` on the command line. You
    can call this function from your own build script to run fbs as if it were
    called via the above command. For an example, see:
        https://build-system.fman.io/manual/#custom-commands
    """
    if project_dir is None:
        project_dir = getcwd()
    try:
        fbs.init(project_dir)
        # Load built-in commands:
        from fbs import builtin_commands
        fn, args = _parse_cmdline()
        fn(*args)
    except KeyboardInterrupt:
        print('')
        sys.exit(-1)
    except FbsError as e:
        # Don't print a stack trace for FbsErrors, just their message:
        _LOG.error(str(e))
        sys.exit(-1)

def command(f):
    """
    Use this as a decorator to define custom fbs commands. For an example, see:
        https://build-system.fman.io/manual/#custom-commands
    """
    COMMANDS[f.__name__] = f
    return f

def _parse_cmdline():
    parser = _get_cmdline_parser()
    args = parser.parse_args()
    if hasattr(args, 'fn'):
        fn_args = []
        for arg in args.args[:1-len(args.defaults)]:
            fn_args.append(getattr(args, arg))
        for arg, default in zip(args.args[-len(args.defaults):], args.defaults):
            fn_args.append(getattr(args, arg, default))
        return args.fn, fn_args
    return parser.print_help, ()

def _get_cmdline_parser():
    # Were we invoked with `python -m fbs`?
    is_python_m_fbs = splitext(basename(sys.argv[0]))[0] == '__main__'
    if is_python_m_fbs:
        prog = '%s -m fbs' % basename(sys.executable)
    else:
        prog = None
    parser = ArgumentParser(prog=prog, description='fbs')
    subparsers = parser.add_subparsers()
    for cmd_name, cmd_fn in COMMANDS.items():
        cmd_parser = subparsers.add_parser(cmd_name, help=cmd_fn.__doc__)
        argspec = getfullargspec(cmd_fn)
        args = argspec.args or []
        defaults = argspec.defaults or ()
        args_without_defaults = args[:1-len(defaults)]
        args_with_defaults = args[-len(defaults):]
        for arg in args_without_defaults:
            cmd_parser.add_argument(arg)
        for arg, default in zip(args_with_defaults, defaults):
            if isinstance(default, bool):
                cmd_parser.add_argument(
                    '--' + arg, action='store_' + str(not default).lower()
                )
            else:
                type_ = None if default is None else type(default)
                cmd_parser.add_argument(arg, default=default, type=type_)
        cmd_parser.set_defaults(fn=cmd_fn, args=args, defaults=defaults)
    return parser


"""
This INTERNAL module is used to manage fbs_runtime's global state. Having it
here, in one central place, allows fbs's test suite to manipulate the state to
simulate various scenarios such as different operating systems.
"""

PLATFORM_NAME = None
LINUX_DISTRIBUTION = None
APPLICATION_CONTEXT = None

def get():
    return PLATFORM_NAME, LINUX_DISTRIBUTION, APPLICATION_CONTEXT

def restore(platform_name, linux_distribution, application_context):
    global PLATFORM_NAME, LINUX_DISTRIBUTION, APPLICATION_CONTEXT
    PLATFORM_NAME = platform_name
    LINUX_DISTRIBUTION = linux_distribution
    APPLICATION_CONTEXT = application_context


"""
This module contains functions that should only be called when running the
frozen form of the app.
"""

from os.path import dirname, join, pardir
from fbs_runtime.platform import is_mac

import sys

# The contents of this dictionary are injected via a PyInstaller runtime hook.
# See: `fbs.freeze._PyInstallerRuntimehook`.
BUILD_SETTINGS = {}

def get_resource_dirs():
    app_dir = dirname(sys.executable)
    return [join(app_dir, pardir, 'Resources') if is_mac() else app_dir]

def load_build_settings():
    return BUILD_SETTINGS


import platform
from fbs import path, SETTINGS
from fbs_runtime.platform import is_mac
from os import replace, remove
from os.path import join, dirname, exists
from subprocess import check_call, DEVNULL


def create_installer_mac():
    app_name = SETTINGS['app_name']
    dest = path('target/${installer}')
    dest_existed = exists(dest)
    if dest_existed:
        dest_bu = dest + '.bu'
        replace(dest, dest_bu)
    try:
        pdata = [
            join(dirname(__file__), 'create-dmg', 'create-dmg'),
            '--volname', app_name,
            '--app-drop-link', '170', '10',
            '--icon', app_name + '.app', '0', '10',
            dest,
            path('${freeze_dir}')
        ]

        if is_mac():
            major, minor = platform.mac_ver()[0].split('.')[:2]
            if (int(major) == 10 and int(minor) >= 15) or int(major) >= 11:
                pdata.insert(1, '--no-internet-enable')

        check_call(pdata, stdout=DEVNULL)
    except:
        if dest_existed:
            replace(dest_bu, dest)
        raise
    else:
        if dest_existed:
            remove(dest_bu)



from fbs_runtime import _state, FbsError

import os
import sys

def is_windows():
    """
    Return True if the current OS is Windows, False otherwise.
    """
    return name() == 'Windows'

def is_mac():
    """
    Return True if the current OS is macOS, False otherwise.
    """
    return name() == 'Mac'

def is_linux():
    """
    Return True if the current OS is Linux, False otherwise.
    """
    return name() == 'Linux'

def name():
    """
    Returns 'Windows', 'Mac' or 'Linux', depending on the current OS. If the OS
    can't be determined, FbsError is raised.
    """
    if _state.PLATFORM_NAME is None:
        _state.PLATFORM_NAME = _get_name()
    return _state.PLATFORM_NAME

def _get_name():
    if sys.platform in ('win32', 'cygwin'):
        return 'Windows'
    if sys.platform == 'darwin':
        return 'Mac'
    if sys.platform.startswith('linux'):
        return 'Linux'
    raise FbsError('Unknown operating system.')

def is_ubuntu():
    try:
        return linux_distribution() in ('Ubuntu', 'Linux Mint')
    except FileNotFoundError:
        return False

def is_arch_linux():
    try:
        return linux_distribution() in ('Arch Linux', 'Manjaro Linux')
    except FileNotFoundError:
        return False

def is_fedora():
    try:
        return linux_distribution() in ('Fedora', 'CentOS Linux')
    except FileNotFoundError:
        return False

def linux_distribution():
    if _state.LINUX_DISTRIBUTION is None:
        _state.LINUX_DISTRIBUTION = _get_linux_distribution()
    return _state.LINUX_DISTRIBUTION

def _get_linux_distribution():
    if not is_linux():
        return ''
    try:
        os_release = _get_os_release_name()
    except OSError:
        pass
    else:
        if os_release:
            return os_release
    return '<unknown>'

def is_gnome_based():
    curr_desktop = os.environ.get('XDG_CURRENT_DESKTOP', '').lower()
    return curr_desktop in ('unity', 'gnome', 'x-cinnamon')

def is_kde_based():
    curr_desktop = os.environ.get('XDG_CURRENT_DESKTOP', '').lower()
    if curr_desktop == 'kde':
        return True
    gdmsession = os.environ.get('GDMSESSION', '').lower()
    return gdmsession.startswith('kde')

def _get_os_release_name():
    with open('/etc/os-release', 'r') as f:
        for line in f:
            line = line.rstrip()
            if line.startswith('NAME='):
                name = line[len('NAME='):]
                return name.strip('"')



from os.path import join, exists, realpath

import errno
import os

class ResourceLocator:
    def __init__(self, resource_dirs):
        self._dirs = resource_dirs
    def locate(self, *rel_path):
        for resource_dir in self._dirs:
            resource_path = join(resource_dir, *rel_path)
            if exists(resource_path):
                return realpath(resource_path)
        raise FileNotFoundError(
            errno.ENOENT, 'Could not locate resource', os.sep.join(rel_path)
        )


from fbs import path, SETTINGS
from fbs_runtime import FbsError
from fbs._state import LOADED_PROFILES
from glob import glob
from os import makedirs
from os.path import dirname, isfile, join, basename, relpath, splitext, exists
from pathlib import Path
from shutil import copy, copymode

import re
import os

def copy_with_filtering(
    src_dir_or_file, dest_dir, replacements=None, files_to_filter=None,
    exclude=None, placeholder='${%s}'
):
    """
    Copy the given file or directory to the given destination, optionally
    applying filtering.
    """
    if replacements is None:
        replacements = SETTINGS
    if files_to_filter is None:
        files_to_filter = []
    if exclude is None:
        exclude = []
    to_copy = _get_files_to_copy(src_dir_or_file, dest_dir, exclude)
    to_filter = _paths(files_to_filter)
    for src, dest in to_copy:
        makedirs(dirname(dest), exist_ok=True)
        if files_to_filter is None or src in to_filter:
            _copy_with_filtering(src, dest, replacements, placeholder)
        else:
            copy(src, dest)

def get_icons():
    """
    Return a list [(size, scale, path)] of available app icons for the current
    platform.
    """
    result = {}
    for profile in LOADED_PROFILES:
        icons_dir = 'src/main/icons/' + profile
        for icon_path in glob(path(icons_dir + '/*.png')):
            name = splitext(basename(icon_path))[0]
            match = re.match('(\d+)(?:@(\d+)x)?', name)
            if not match:
                raise FbsError('Invalid icon name: ' + icon_path)
            size, scale = int(match.group(1)), int(match.group(2) or '1')
            result[(size, scale)] = icon_path
    return [(size, scale, path) for (size, scale), path in result.items()]

def _get_files_to_copy(src_dir_or_file, dest_dir, exclude):
    excludes = _paths(map(path, exclude))
    if isfile(src_dir_or_file) and src_dir_or_file not in excludes:
        yield src_dir_or_file, join(dest_dir, basename(src_dir_or_file))
    else:
        for (subdir, _, files) in os.walk(src_dir_or_file):
            dest_subdir = join(dest_dir, relpath(subdir, src_dir_or_file))
            for file_ in files:
                file_path = join(subdir, file_)
                dest_path = join(dest_subdir, file_)
                if file_path not in excludes:
                    yield file_path, dest_path

def _copy_with_filtering(
    src_file, dest_file, dict_, placeholder='${%s}', encoding='utf-8'
):
    replacements = []
    for key, value in dict_.items():
        old = (placeholder % key).encode(encoding)
        new = str(value).encode(encoding)
        replacements.append((old, new))
    with open(src_file, 'rb') as open_src_file:
        with open(dest_file, 'wb') as open_dest_file:
            for line in open_src_file:
                new_line = line
                for old, new in replacements:
                    new_line = new_line.replace(old, new)
                open_dest_file.write(new_line)
        copymode(src_file, dest_file)

class _paths:
    def __init__(self, paths):
        self._paths = []
        # _defaults includes "files_to_filter" - eg. Installer.nsi. If these
        # files don't also exist in the "user's" src/ directory, then
        # Path(p).resolve() raises FileNotFoundError. Handle this:
        for p in paths:
            try:
                self._paths.append(self._resolve_strict(Path(p)))
            except FileNotFoundError:
                pass
    def __contains__(self, item):
        item = Path(item).resolve()
        for p in self._paths:
            # We do `p == item` here instead of `p.samefile(item)` because a
            # user reported that the former does not work. The affected paths
            # were in a VirtualBox shared folder in a Windows guest. They had
            # different st_ino values even though the paths were the same.
            # See https://github.com/mherrmann/fbs/issues/112.
            if p == item or p in item.parents:
                return True
        return False
    def _resolve_strict(self, path_):
        try:
            return path_.resolve(strict=True)
        except TypeError:
            # Python < 3.6:
            return path_.resolve()

def _copy(path_fn, src, dst): # Used by several other internal fbs modules
    src = path_fn(src)
    if exists(src):
        filter_ = [path_fn(f) for f in SETTINGS['files_to_filter']]
        copy_with_filtering(src, dst, files_to_filter=filter_)
        return True
    return False


"""
By default, fbs overwrites sys.excepthook for better error reporting:

 1) Applications based on PyQt5 or PySide2 are missing some stack trace entries
    in their tracebacks (see add_missing_qt_frames(...) below). fbs makes sure
    they are displayed, for easier debugging.
 2) Python < 3.8 does not call sys.excepthook for non-main threads. fbs ensures
    that its own excepthook (and thus eg. the benefits of 1) above) does get
    called, so you see all errors.
 3) fbs lets you define multiple `ExceptionHandler`s - see below. This lets you
    report errors through several channels. Eg. on the console, or on a hosted
    error reporting system.

You can customise these mechanisms using the classes and functions in this
module, and by changing ApplicationContext#exception_handlers and
...#excepthook.
"""

from collections import namedtuple

import sys
import threading
import traceback

class ExceptionHandler:
    """
    Extend this class to implement your own exception handler(s). Then, add it
    to your ApplicationContext#exception_handlers.
    """
    def init(self):
        pass
    def handle(self, exc_type, exc_value, enriched_tb):
        """
        Return True from this method to prevent further ExceptionHandlers from
        being invoked for this exception.
        """
        raise NotImplementedError()

class StderrExceptionHandler(ExceptionHandler):
    """
    Print exceptions to stderr.
    """
    def handle(self, exc_type, exc_value, enriched_tb):
        # Normally, we'd like to use sys.__excepthook__ here. But it doesn't
        # work with our "fake" traceback (see add_missing_qt_frames(...)).
        # The following call avoids this yet produces the same result:
        traceback.print_exception(exc_type, exc_value, enriched_tb)

class _Excepthook:
    """
    fbs's excepthook. Forwards exceptions to the given handlers, until one of
    them returns True. Adds stack trace entries that are normally missing in
    PyQt5 / PySide2 applications (see add_missing_qt_frames(...)). Also ensures
    that, unlike in Python normally, it is called for exceptions in all threads.
    """
    def __init__(self, handlers):
        self._handlers = handlers
    def install(self):
        for handler in self._handlers:
            handler.init()
        sys.excepthook = self
        enable_excepthook_for_threads()
    def __call__(self, exc_type, exc_value, exc_tb):
        if not isinstance(exc_value, SystemExit):
            enriched_tb = add_missing_qt_frames(exc_tb) if exc_tb else exc_tb
            for handler in self._handlers:
                if handler.handle(exc_type, exc_value, enriched_tb):
                    break

def enable_excepthook_for_threads():
    """
    `sys.excepthook` isn't called for exceptions raised in non-main-threads.
    This workaround fixes this for instances of (non-subclasses of) Thread.
    See: http://bugs.python.org/issue1230540
    """
    init_original = threading.Thread.__init__

    def init(self, *args, **kwargs):
        init_original(self, *args, **kwargs)
        run_original = self.run

        def run_with_except_hook(*args2, **kwargs2):
            try:
                run_original(*args2, **kwargs2)
            except Exception:
                sys.excepthook(*sys.exc_info())

        self.run = run_with_except_hook

    threading.Thread.__init__ = init

def add_missing_qt_frames(tb):
    """
    Let f and h be Python functions and g be a function of Qt. If
        f() -> g() -> h()
    (where "->" means "calls"), and an exception occurs in h(), then the
    traceback does not contain f. This can make debugging very difficult.
    To fix this, this function creates a "fake" traceback that contains the
    missing entries.

    The code below can be used to reproduce the f() -> g() -> h() problem.
    It opens a window with a button. When you click it, an error occurs
    whose traceback does not include f().

    The problem described here is not specific to PyQt5 - It occurs for
    PySide2 as well. To see this, replace PyQt5 by PySide2 below.

        from PyQt5.QtWidgets import *
        from PyQt5.QtCore import Qt

        class Window(QWidget):
            def __init__(self):
                super().__init__()
                btn = QPushButton('Click me', self)
                btn.clicked.connect(self.f)
            def f(self, _):
                self.inputMethodQuery(Qt.ImAnchorPosition)
            def inputMethodQuery(self, query):
                if query == Qt.ImAnchorPosition:
                    # Make Qt call inputMethodQuery(ImCursorPosition).
                    # This is our "g()":
                    return super().inputMethodQuery(query) # "g()"
                self.h()
            def h(self):
                raise Exception()

        app = QApplication([])
        window = Window()
        window.show()
        app.exec_()
    """
    result = _fake_tb(tb.tb_frame, tb.tb_lasti, tb.tb_lineno, tb.tb_next)
    frame = tb.tb_frame.f_back
    while frame:
        result = _fake_tb(frame, frame.f_lasti, frame.f_lineno, result)
        frame = frame.f_back
    return result

_fake_tb = \
    namedtuple('fake_tb', ('tb_frame', 'tb_lasti', 'tb_lineno', 'tb_next'))


from collections import namedtuple
from fbs_runtime import _state, _frozen, _source
from fbs_runtime._resources import ResourceLocator
from fbs_runtime._signal import SignalWakeupHandler
from fbs_runtime.excepthook import _Excepthook, StderrExceptionHandler
from fbs_runtime.platform import is_windows, is_mac
from functools import lru_cache

import sys

def cached_property(getter):
    """
    A cached Python @property. You use it in conjunction with ApplicationContext
    below to instantiate the components that comprise your application. For more
    information, please consult the Manual:
        https://build-system.fman.io/manual/#cached_property
    """
    return property(lru_cache()(getter))

class _ApplicationContext:
    """
    The main point of contact between your application and fbs. For information
    on how to use it, please see the Manual:
        https://build-system.fman.io/manual/#your-python-code
    """
    def __init__(self):
        if self.excepthook:
            self.excepthook.install()
        # Many Qt classes require a QApplication to have been instantiated.
        # Do this here, before everything else, to achieve this:
        self.app
        # We don't build as a console app on Windows, so no point in installing
        # the SIGINT handler:
        if not is_windows():
            self._signal_wakeup_handler = \
                SignalWakeupHandler(self.app, self._qt_binding.QAbstractSocket)
            self._signal_wakeup_handler.install()
        if self.app_icon:
            self.app.setWindowIcon(self.app_icon)
    def run(self):
        """
        You should overwrite this method with the steps for starting your app.
        See eg. fbs's tutorial.
        """
        raise NotImplementedError()
    @cached_property
    def app(self):
        """
        The global Qt QApplication object for your app. Feel free to overwrite
        this property, eg. if you wish to use your own subclass of QApplication.
        An example of this is given in the Manual.
        """
        result = self._qt_binding.QApplication([])
        result.setApplicationName(self.build_settings['app_name'])
        result.setApplicationVersion(self.build_settings['version'])
        return result
    @cached_property
    def build_settings(self):
        """
        This dictionary contains the values of the settings listed in setting
        "public_settings". Eg. `self.build_settings['version']`.
        """
        if is_frozen():
            return _frozen.load_build_settings()
        return _source.load_build_settings(self._project_dir)
    def get_resource(self, *rel_path):
        """
        Return the absolute path to the data file with the given name or
        (relative) path. When running from source, searches src/main/resources.
        Otherwise, searches your app's installation directory. If no file with
        the given name or path exists, a FileNotFoundError is raised.
        """
        return self._resource_locator.locate(*rel_path)
    @cached_property
    def exception_handlers(self):
        """
        Return a list of exception handlers that should be invoked when an error
        occurs. See the documentation of module `fbs_runtime.excepthook` for
        more information.
        """
        return [StderrExceptionHandler()]
    @cached_property
    def licensing(self):
        """
        This field helps you implement a license key functionality for your
        application. For more information, see:
            https://build-system.fman.io/manual#license-keys
        """

        # fbs's licensing implementation incurs a dependency on Python library
        # `rsa`. We don't want to force all users to install this library.
        # So we import fbs_runtime.licensing here, instead of at the top of this
        # file. This lets people who don't use licensing avoid the dependency.
        from fbs_runtime.licensing import _Licensing

        return _Licensing(self.build_settings['licensing_pubkey'])
    @cached_property
    def app_icon(self):
        """
        The app icon. Not available on Mac because app icons are handled by the
        OS there.
        """
        if not is_mac():
            return self._qt_binding.QIcon(self.get_resource('Icon.ico'))
    @cached_property
    def excepthook(self):
        """
        Overwrite this method to use a custom excepthook. It should be an object
        with a .install() method, or `None` if you want to completely disable
        fbs's excepthook implementation.
        """
        return _Excepthook(self.exception_handlers)
    @cached_property
    def _qt_binding(self):
        # Implemented in subclasses.
        raise NotImplementedError()
    @cached_property
    def _resource_locator(self):
        if is_frozen():
            resource_dirs = _frozen.get_resource_dirs()
        else:
            resource_dirs = _source.get_resource_dirs(self._project_dir)
        return ResourceLocator(resource_dirs)
    @cached_property
    def _project_dir(self):
        assert not is_frozen(), 'Only available when running from source'
        return _source.get_project_dir()

_QtBinding = \
    namedtuple('_QtBinding', ('QApplication', 'QIcon', 'QAbstractSocket'))

def is_frozen():
    """
    Return True if running from the frozen (i.e. compiled form) of your app, or
    False when running from source.
    """
    return getattr(sys, 'frozen', False)

def get_application_context(DevelopmentAppCtxtCls, FrozenAppCtxtCls=None):
    if FrozenAppCtxtCls is None:
        FrozenAppCtxtCls = DevelopmentAppCtxtCls
    if _state.APPLICATION_CONTEXT is None:
        _state.APPLICATION_CONTEXT = \
            FrozenAppCtxtCls() if is_frozen() else DevelopmentAppCtxtCls()
    return _state.APPLICATION_CONTEXT


from fbs import path, SETTINGS
from fbs.installer import _generate_installer_resources
from fbs.resources import get_icons
from fbs_runtime.platform import is_arch_linux
from os import makedirs, remove, rename
from os.path import join, dirname, exists
from shutil import copy, rmtree, copytree
from subprocess import run, DEVNULL

def generate_installer_files():
    if exists(path('target/installer')):
        rmtree(path('target/installer'))
    copytree(path('${freeze_dir}'), path('target/installer/opt/${app_name}'))
    _generate_installer_resources()
    # Special handling of the .desktop file: Replace AppName by actual name.
    apps_dir = path('target/installer/usr/share/applications')
    rename(
        join(apps_dir, 'AppName.desktop'),
        join(apps_dir, SETTINGS['app_name'] + '.desktop')
    )
    _generate_icons()

def run_fpm(output_type):
    dest = path('target/${installer}')
    if exists(dest):
        remove(dest)
    # Lower-case the name to avoid the following fpm warning:
    #  > Debian tools (dpkg/apt) don't do well with packages that use capital
    #  > letters in the name. In some cases it will automatically downcase
    #  > them, in others it will not. It is confusing. Best to not use any
    #  > capital letters at all.
    name = SETTINGS['app_name'].lower()
    args = [
        'fpm', '-s', 'dir',
        # We set the log level to error because fpm prints the following warning
        # even if we don't have anything in /etc:
        #  > Debian packaging tools generally labels all files in /etc as config
        #  > files, as mandated by policy, so fpm defaults to this behavior for
        #  > deb packages. You can disable this default behavior with
        #  > --deb-no-default-config-files flag
        '--log', 'error',
        '-C', path('target/installer'),
        '-n', name,
        '-v', SETTINGS['version'],
        '--vendor', SETTINGS['author'],
        '-t', output_type,
        '-p', dest
    ]
    if SETTINGS['description']:
        args.extend(['--description', SETTINGS['description']])
    if SETTINGS['author_email']:
        args.extend([
            '-m', '%s <%s>' % (SETTINGS['author'], SETTINGS['author_email'])
        ])
    if SETTINGS['url']:
        args.extend(['--url', SETTINGS['url']])
    for dependency in SETTINGS['depends']:
        args.extend(['-d', dependency])
    if is_arch_linux():
        for opt_dependency in SETTINGS['depends_opt']:
            args.extend(['--pacman-optional-depends', opt_dependency])
    try:
        run(args, check=True, stdout=DEVNULL)
    except FileNotFoundError:
        raise FileNotFoundError(
            "fbs could not find executable 'fpm'. Please install fpm using the "
            "instructions at "
            "https://fpm.readthedocs.io/en/latest/installing.html."
        ) from None

def _generate_icons():
    dest_root = path('target/installer/usr/share/icons/hicolor')
    makedirs(dest_root)
    icons_fname = '%s.png' % SETTINGS['app_name']
    for size, _, icon_path in get_icons():
        icon_dest = join(dest_root, '%dx%d' % (size, size), 'apps', icons_fname)
        makedirs(dirname(icon_dest))
        copy(icon_path, icon_dest)


from logging import StreamHandler
from textwrap import wrap

import fbs.cmdline
import logging
import sys

def _main():
    """
    Main entry point for the `fbs` command line script.

    We init logging here instead of in fbs.cmdline.main(...) because the latter
    can be called by projects using fbs, and it's bad practice for libraries to
    configure logging. See eg. https://stackoverflow.com/a/26087972/1839209.
    """
    _init_logging()
    fbs.cmdline.main()

def _init_logging():
    # Redirect INFO or lower to stdout, WARNING or higher to stderr:
    stdout = _WrappingStreamHandler(sys.stdout)
    stdout.setLevel(logging.DEBUG)
    stdout.addFilter(lambda record: record.levelno <= logging.INFO)
    # Don't wrap stderr because it may contain stack traces:
    stderr = logging.StreamHandler(sys.stderr)
    stderr.setLevel(logging.WARNING)
    logging.basicConfig(
        level=logging.INFO, format='%(message)s', handlers=(stdout, stderr)
    )

class _WrappingStreamHandler(StreamHandler):
    def __init__(self, stream=None, line_length=70):
        super().__init__(stream)
        self._line_length = line_length
    def format(self, record):
        result = super().format(record)
        if not getattr(record, 'wrap', True):
            # Make it possible to prevent wrapping. Eg.:
            #     _LOG.info('Message', extra={'wrap': False})
            return result
        lines = result.split(self.terminator)
        new_lines = []
        for line in lines:
            new_lines.extend(
                wrap(line, self._line_length, replace_whitespace=False) or ['']
            )
        return self.terminator.join(new_lines)

if __name__ == '__main__':
    _main()


"""
Earlier fbs versions had the following code:

    try:
        from PyQt5 import ...
    except ImportError:
        from PySide2 import ...

This lead to problems when both PyQt5 and PySide2 were on PYTHONPATH:

 1) PyInstaller packaged both (!) libraries because it saw both imports.
 2) The above made fbs always use PyQt5. But if the user's app uses PySide2,
    then PySide2 and PyQt5 classes / code would be mixed.
 3) It wasn't clear (or deterministic, really) which Python binding took
    precedence. For instance, PyQt5 and PySide2 set different QML search paths.

To fix this problems, the above code was split into separate files: One that
contains all PyQt5 imports, and another that contains all PySide2 imports. The
user is supposed to import precisely one of the two. This makes PyInstaller
only package the one necessary library, and prevents the above problems.
"""

from . import _ApplicationContext, _QtBinding, cached_property
from PySide2.QtGui import QIcon
from PySide2.QtWidgets import QApplication
from PySide2.QtNetwork import QAbstractSocket

class ApplicationContext(_ApplicationContext):
    @cached_property
    def _qt_binding(self):
        return _QtBinding(QApplication, QIcon, QAbstractSocket)


from base64 import b64encode, b64decode
from rsa import PrivateKey, PublicKey, VerificationError

import json
import rsa

def pack_license_key(data, privkey_args):
    """
    Pack a dictionary of license key data to a string. You typically call this
    function on a server, when a user purchases a license. Eg.:

        lk_contents = pack_license_key({'email': 'some@user.com'}, ...)

    The parameter `privkey_args` is a dictionary containing values for the RSA
    fields "n", "e", "d", "p" and "q". You can generate it with fbs's command
    `init_licensing`.

    The resulting string is signed to prevent the end user from changing it.
    Use the function `unpack_license_key` below to reconstruct `data` from it.
    This also verifies that the string was not tampered with.

    This function has two non-obvious caveats:

    1) It does not obfuscate the data. If `data` contains "key": "value", then
       "key": "value" is also visible in the resulting string.

    2) Calling this function twice with the same arguments will result in the
    same string. This may be undesirable when you generate multiple license keys
    for the same user. A simple workaround for this is to add a unique parameter
    to `data`, such as the current timestamp.
    """
    data_bytes = _dumpb(data)
    signature = rsa.sign(data_bytes, PrivateKey(**privkey_args), 'SHA-1')
    result = dict(data)
    if 'key' in data:
        raise ValueError('Data must not contain an element called "key"')
    result['key'] = b64encode(signature).decode('ascii')
    return json.dumps(result)

class _Licensing:
    """
    This internal class lets us inject the licensing functionality into the
    application context, in such a way that the fbs user does not have to worry
    about where the public key is stored.
    """
    def __init__(self, pubkey_args):
        self._pubkey = pubkey_args
    def unpack_license_key(self, key_str):
        return unpack_license_key(key_str, self._pubkey)

class InvalidKey(Exception):
    pass

def unpack_license_key(key_str, pubkey_args):
    """
    Decode a string of license key data produced by `pack_license_key`. In other
    words, this function is the inverse of `pack_...` above:

        data == unpack_license_key(pack_license_key(data, ...), ...)

    If the given string is not a valid key, `InvalidKey` is raised.

    The parameter `pubkey_args` is a dictionary containing values for the RSA
    fields "n" and "e". It can be generated with fbs's command `init_licensing`.
    """
    try:
        result = json.loads(key_str)
    except ValueError:
        raise InvalidKey() from None
    try:
        signature = result.pop('key')
    except KeyError:
        raise InvalidKey() from None
    try:
        signature_bytes = b64decode(signature.encode('ascii'))
    except ValueError:
        raise InvalidKey() from None
    try:
        rsa.verify(_dumpb(result), signature_bytes, PublicKey(**pubkey_args))
    except VerificationError:
        raise InvalidKey() from None
    return result

def _dumpb(dict_):
    return json.dumps(dict_, sort_keys=True).encode('utf-8')


from fbs import path
from fbs.installer.linux import generate_installer_files, run_fpm
from subprocess import run

def create_installer_arch():
    generate_installer_files()
    # Avoid pacman warning "directory permissions differ" when installing:
    run(['chmod', 'g-w', '-R', path('target/installer')], check=True)
    run_fpm('pacman')


from fbs_runtime.application_context.${python_bindings} import ApplicationContext
from ${python_bindings}.QtWidgets import QMainWindow

import sys

if __name__ == '__main__':
    appctxt = ApplicationContext()       # 1. Instantiate ApplicationContext
    window = QMainWindow()
    window.resize(250, 150)
    window.show()
    exit_code = appctxt.app.exec_()      # 2. Invoke appctxt.app.exec_()
    sys.exit(exit_code)


from argparse import ArgumentParser
from os import makedirs
from os.path import join, relpath, dirname, isdir
from shutil import copyfile

import hashlib
import os

_BASE_FILE_NAME = 'base'

class UnknownHash(Exception):
    pass

def pack(main_py, src_dir, dst_dir):
    """
    On the first call, copies `src_dir` into `dst_dir`/<hash>/ and creates
    `dst_dir`/base with contents <hash>, where <hash> is the hash of `main_py`.

    On the second call, assuming <hash> is different, copies only those files
    that are new or changed into `dst_dir`/<hash>/.

    The motivation is the following: We want to pack directories whose contents
    are largely the same. Eg.:

        1/
            MyApp/
        2/
            MyApp/

    99% of the files in 1/MyApp and 2/MyApp have the same contents.
    Unfortunately, Python's shutil.make_archive(...) produces an archive that is
    almost twice the size of what we get when we compress just one of the two.
    By skipping unchanged files, the present implementation achieves
    significantly better compression.
    """
    main_py_hash = get_py_file_hash(main_py)
    base_file = join(dst_dir, _BASE_FILE_NAME)
    try:
        with open(base_file) as f:
            base = f.read()
    except FileNotFoundError:
        base = main_py_hash
        makedirs(dirname(base_file), exist_ok=True)
        with open(base_file, 'w') as f:
            f.write(main_py_hash)
    for subdir, _, files in os.walk(src_dir):
        for file_name in files:
            file_path = join(subdir, file_name)
            file_rel_path = relpath(file_path, src_dir)
            base_copy = join(dst_dir, base, file_rel_path)
            with open(file_path, 'rb') as f2:
                try:
                    with open(base_copy, 'rb') as f:
                        is_equal = f.read() == f2.read()
                except FileNotFoundError:
                    is_equal = False
            if not is_equal:
                dst_path = join(dst_dir, main_py_hash, file_rel_path)
                makedirs(dirname(dst_path), exist_ok=True)
                copyfile(file_path, dst_path)

def unpack(main_py, src_dir, dst_dir):
    main_py_hash = get_py_file_hash(main_py)
    if not isdir(join(src_dir, main_py_hash)):
        raise UnknownHash()
    base_file = join(src_dir, _BASE_FILE_NAME)
    with open(base_file) as f:
        base = f.read()
    _copytree(join(src_dir, base), dst_dir)
    _copytree(join(src_dir, main_py_hash), dst_dir)

def get_py_file_hash(file_path):
    with open(file_path) as f:
        contents_no_newlines = ''.join(line.strip() for line in f)
    m = hashlib.sha1()
    m.update(contents_no_newlines.encode('utf-8'))
    return m.digest().hex()

def _copytree(src_dir, dst_dir):
    for subdir, _, files in os.walk(src_dir):
        for file_name in files:
            file_path = join(subdir, file_name)
            file_rel_path = relpath(file_path, src_dir)
            dst_path = join(dst_dir, file_rel_path)
            makedirs(dirname(dst_path), exist_ok=True)
            copyfile(file_path, dst_path)

if __name__ == '__main__':
    parser = ArgumentParser(description='Package files for an OS shim.')
    parser.add_argument('main', help='main.py script')
    parser.add_argument('src', help='source dir. Eg. .../target/ubuntu/MyApp')
    parser.add_argument(
        'dst', help='dest dir. Eg. .../fbs_tutorial_shim_ubuntu/data'
    )
    args = parser.parse_args()
    pack(args.main, args.src, args.dst)


"""
This module contains functions that should only be called by module `fbs`, or
when running from source.
"""

from fbs_runtime import FbsError
from fbs_runtime._fbs import get_default_profiles, get_core_settings, \
    filter_public_settings
from fbs_runtime._settings import load_settings
from os.path import join, normpath, dirname, pardir, exists
from pathlib import Path

import os

def get_project_dir():
    result = Path(os.getcwd())
    while result != result.parent:
        if (result / 'src' / 'main' / 'python').is_dir():
            return str(result)
        result = result.parent
    raise FbsError(
        'Could not determine the project base directory. '
        'Was expecting src/main/python.'
    )

def get_resource_dirs(project_dir):
    resources = path(project_dir, 'src/main/resources')
    result = [join(resources, profile) for profile in get_default_profiles()]
    result.append(path(project_dir, 'src/main/icons'))
    return result

def load_build_settings(project_dir):
    core_settings = get_core_settings(project_dir)
    profiles = get_default_profiles()
    json_paths = get_settings_paths(project_dir, profiles)
    all_settings = load_settings(json_paths, core_settings)
    return filter_public_settings(all_settings)

def get_settings_paths(project_dir, profiles):
    return list(filter(exists, (
        path_fn('src/build/settings/%s.json' % profile)
        for path_fn in (default_path, lambda p: path(project_dir, p))
        for profile in profiles
    )))

def default_path(path_str):
    defaults_dir = join(dirname(__file__), pardir, 'fbs', '_defaults')
    return path(defaults_dir, path_str)

def path(base_dir, path_str):
    return normpath(join(base_dir, *path_str.split('/')))


from fbs_runtime import platform
from fbs_runtime.platform import is_ubuntu, is_linux, is_arch_linux, is_fedora

def get_core_settings(project_dir):
    return {
        'project_dir': project_dir
    }

def get_default_profiles():
    result = ['base']
    # The "secret" profile lets the user store sensitive settings such as
    # passwords in src/build/settings/secret.json. When using Git, the user can
    # exploit this by adding secret.json to .gitignore, thus preventing it from
    # being uploaded to services such as GitHub.
    result.append('secret')
    result.append(platform.name().lower())
    if is_linux():
        if is_ubuntu():
            result.append('ubuntu')
        elif is_arch_linux():
            result.append('arch')
        elif is_fedora():
            result.append('fedora')
    return result

def filter_public_settings(settings):
    return {k: settings[k] for k in settings['public_settings']}


from socket import socketpair, SOCK_DGRAM

import signal

class SignalWakeupHandler:
    """
    Python's `signal` module lets us define custom signal handlers. What we want
    in particular is a graceful handling of Ctrl+C, meaning that the app shuts
    down cleanly.

    The default implementation in Python (ie. if we don't set any signals) is to
    raise KeyboardInterrupt at arbitrary points in our code. This is not nice or
    easy to handle.

    The simplest implementation would be to call `signal(SIGINT, SIG_DFL)`. This
    immediately kills the app when the user presses Ctrl in the console window
    attached to it. This is not nice because no shutdown/cleanup code is
    executed.

    Another implementation would be to call `signal(SIGINT, ...app.exit())`. The
    problem is: This signal seems to only be delivered when the app has focus.
    So the user presses Ctrl+C in the console, then needs to switch to the GUI
    window for the signal to be delivered and the app to shut down. Not nice
    either.

    The present class fixes this problem. It integrates Python and Qt so that
    signal handlers installed with Python get called immediately. This way, we
    can immediately quit the app when the user presses Ctrl+C in the console.

    See: https://stackoverflow.com/a/37229299/1839209
    """
    def __init__(self, app, QAbstractSocket):
        self._app = app
        self.old_fd = None
        # Create a socket pair
        self.wsock, self.rsock = socketpair(type=SOCK_DGRAM)
        self.socket = QAbstractSocket(QAbstractSocket.UdpSocket, app)
        # Let Qt listen on the one end
        self.socket.setSocketDescriptor(self.rsock.fileno())
        # And let Python write on the other end
        self.wsock.setblocking(False)
        self.old_fd = signal.set_wakeup_fd(self.wsock.fileno())
        # First Python code executed gets any exception from
        # the signal handler, so add a dummy handler first
        self.socket.readyRead.connect(lambda : None)
        # Second handler does the real handling
        self.socket.readyRead.connect(self._readSignal)
    def install(self):
        signal.signal(signal.SIGINT, lambda *_: self._app.exit(130))
    def __del__(self):
        # Restore any old handler on deletion
        if self.old_fd is not None and signal and signal.set_wakeup_fd:
            signal.set_wakeup_fd(self.old_fd)
    def _readSignal(self):
        # Read the written byte.
        # Note: readyRead is blocked from occurring again until readData()
        # was called, so call it, even if you don't need the value.
        _ = self.socket.readData(1)


from fbs import path, LOADED_PROFILES
from fbs.resources import _copy
from fbs_runtime._source import default_path

def _generate_installer_resources():
    for path_fn in default_path, path:
        for profile in LOADED_PROFILES:
            _copy(path_fn, 'src/installer/' + profile, path('target/installer'))


"""
This INTERNAL module is used to manage fbs's global state. Having it here, in
one central place, allows fbs's test suite to manipulate the state to test
various scenarios.
"""
from collections import OrderedDict

SETTINGS = {}
LOADED_PROFILES = []
COMMANDS = OrderedDict()

def get():
    return dict(SETTINGS), list(LOADED_PROFILES), dict(COMMANDS)

def restore(settings, loaded_profiles, commands):
    SETTINGS.clear()
    SETTINGS.update(settings)
    LOADED_PROFILES.clear()
    LOADED_PROFILES.extend(loaded_profiles)
    COMMANDS.clear()
    COMMANDS.update(commands)


from fbs import path
from fbs.installer import _generate_installer_resources
from subprocess import check_call, DEVNULL

def create_installer_windows():
    _generate_installer_resources()
    try:
        check_call(
            ['makensis', 'Installer.nsi'], cwd=path('target/installer'),
            stdout=DEVNULL
        )
    except FileNotFoundError:
        raise FileNotFoundError(
            "fbs could not find executable 'makensis'. Please install NSIS and "
            "add its installation directory to your PATH environment variable."
        ) from None


"""fbs tutorial helper

See:
https://github.com/mherrmann/fbs-tutorial-shim
"""

from os.path import relpath, join
from setuptools import setup, find_packages

import os

def _get_package_data(pkg_dir, data_subdir):
    result = []
    for dirpath, _, filenames in os.walk(join(pkg_dir, data_subdir)):
        for filename in filenames:
            filepath = join(dirpath, filename)
            result.append(relpath(filepath, pkg_dir))
    return result

description = 'fbs tutorial helper'
setup(
    name='fbs-tutorial-shim',
    version='1.0.0',
    description=description,
    long_description=
        description + '\n\nSee: https://github.com/mherrmann/fbs-tutorial-shim',
    author='Michael Herrmann',
    author_email='michael+removethisifyouarehuman@herrmann.io',
    url='https://github.com/mherrmann/fbs-tutorial-shim',
    packages=find_packages(exclude=('tests', 'tests.*')),
    package_data={
        'fbs': _get_package_data('fbs', '_defaults'),
        'fbs.builtin_commands':
            _get_package_data('fbs/builtin_commands', 'project_template')
    },
    install_requires=[
        'PyQt5',
        'fbs-tutorial-shim-windows; sys_platform=="win32"',
        'fbs-tutorial-shim-mac; sys_platform=="darwin"'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
    
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    
        'Operating System :: OS Independent',
    
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',

        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    entry_points={
        'console_scripts': ['fbs=fbs.__main__:_main']
    },
    license='GPLv3 or later',
    keywords='PyQt',
    platforms=['MacOS', 'Windows'],
    test_suite='tests'
)


from collections import OrderedDict
from fbs import path
from fbs_runtime import FbsError
from getpass import getpass
from os.path import exists

BASE_JSON = 'src/build/settings/base.json'
SECRET_JSON = 'src/build/settings/secret.json'

def prompt_for_value(
    value, optional=False, default='', password=False, choices=()
):
    message = value
    if choices:
        choices_dict = \
            OrderedDict((str(i + 1), c) for (i, c) in enumerate(choices))
        message += ': '
        message += ' or '.join('%s) %s' % tpl for tpl in choices_dict.items())
    if default:
        message += ' [%s] ' % \
                   (choices.index(default) + 1 if choices else default)
    message += ': '
    prompt = getpass if password else input
    result = prompt(message).strip()
    if not result and default:
        print(default)
        return default
    if not optional:
        while not result or (choices and result not in choices_dict):
            result = prompt(message).strip()
    return choices_dict[result] if choices else result

def require_existing_project():
    if not exists(path('src')):
        raise FbsError(
            "Could not find the src/ directory. Are you in the right folder?\n"
            "If yes, did you already run\n"
            "    fbs startproject ?"
        )

def require_frozen_app():
    if not exists(path('${freeze_dir}')):
        raise FbsError(
            'It seems your app has not yet been frozen. Please run:\n'
            '    fbs freeze'
        )


from time import time

class RateLimiter:
    def __init__(self, interval_secs, allowance, time_fn=time):
        self._interval = interval_secs
        self._allowance = allowance
        self._time_fn = time_fn
        self._interval_start = time_fn()
        self._num_requests = 0
    def please(self):
        now = self._time_fn()
        if now > self._interval_start + self._interval:
            self._num_requests = 0
            self._interval_start = now
        if self._num_requests < self._allowance:
            self._num_requests += 1
            return True
        return False


import json

def load_settings(json_paths, base=None):
    """
    Return settings from the given JSON files as a dictionary. This function
    expands placeholders: That is, if a settings file contains

        {
            "app_name": "MyApp",
            "freeze_dir": "target/${app_name}"
        }

    then "freeze_dir" in the result of this function is "target/MyApp".

    It also merges lists. Say base.json contains

        { "hidden_imports": ["a"] }

    and mac.json contains

        { "hidden_imports": ["b"] }

    then you obtain

        { "hidden_imports": ["a", "b'] }.
    """
    if base is None:
        result = None
    else:
        result = dict(base)
    for json_path in json_paths:
        with open(json_path, 'r') as f:
            data = json.load(f)
        result = data if result is None else _merge(result, data)
    while True:
        for key, value in result.items():
            new_value = expand_placeholders(value, result)
            if new_value != value:
                result[key] = new_value
                break
        else:
            break
    return result

def expand_placeholders(obj, settings):
    if isinstance(obj, str):
        for key, value in settings.items():
            obj = obj.replace('${%s}' % key, str(value))
    elif isinstance(obj, list):
        return [expand_placeholders(o, settings) for o in obj]
    elif isinstance(obj, dict):
        return {k: expand_placeholders(v, settings) for k, v in obj.items()}
    return obj

def _merge(a, b):
    if type(a) != type(b):
        raise ValueError('Cannot merge %r and %r' % (a, b))
    if isinstance(a, list):
        return a + b
    if isinstance(a, dict):
        result = dict(a)
        for k, v in b.items():
            result[k] = _merge(a[k], v) if k in a else v
        return result
    return b


from fbs_runtime.excepthook import ExceptionHandler
from fbs_runtime.excepthook._util import RateLimiter

import sentry_sdk

class SentryExceptionHandler(ExceptionHandler):
    """
    Send stack traces to Sentry. For instructions on how to set this up, see:

        https://build-system.fman.io/manual/#error-tracking

    A property of interest in this class is .scope: It lets you send additional
    context information to Sentry, such as the user's operating system, or their
    email. These data are then displayed alongside any stack traces in Sentry.

    A limitation of .scope is that it is only available once .init() was called.
    fbs's ApplicationContext performs this call automatically for handlers
    listed in the .exception_handlers property.

    The recommended way for setting context information is to use the `callback`
    parameter. You can see its use near the bottom of the following snippet:

        from fbs_runtime import platform
        from fbs_runtime.application_context import ApplicationContext, \
            cached_property, is_frozen

        class AppContext(ApplicationContext):
            ...
            @cached_property
            def exception_handlers(self):
                result = super().exception_handlers
                if is_frozen():
                    result.append(self.sentry)
                return result
            @cached_property
            def sentry(self):
                # The Sentry client key. Eg. https://4e78a0...@sentry.io/12345.
                dsn = self.build_settings['sentry_dsn']
                # Your app version. Eg. 1.2.3:
                version = self.build_settings['version']
                # The environment in which your app is running. "local" by
                # default, but set to "production" when you do `fbs release`.
                environment = self.build_settings['environment']
                return SentryExceptionHandler(
                    dsn, version, environment, callback=self._on_sentry_init
                )
            def _on_sentry_init(self):
                self.sentry.scope.set_extra('os', platform.name())
                self.sentry.scope.user = {'id': 41, 'email': 'john@gmail.com'}

    The optional `rate_limit` parameter to the constructor lets you limit the
    number of requests per minute. It is there to prevent a single client from
    clogging up your Sentry logs.
    """
    def __init__(
        self, dsn, app_version, environment, callback=lambda: None,
        rate_limit=10
    ):
        super().__init__()
        self.scope = None
        self._dsn = dsn
        self._app_version = app_version
        self._environment = environment
        self._callback = callback
        self._rate_limiter = RateLimiter(60, rate_limit)
    def init(self):
        sentry_sdk.init(
            self._dsn, release=self._app_version, environment=self._environment,
            attach_stacktrace=True, default_integrations=False
        )
        # Sentry doesn't give us an easy way to set context information
        # globally, for all threads. We work around this by maintaining a
        # reference to "the" main scope:
        self.scope = sentry_sdk.configure_scope().__enter__()
        self._callback()
    def handle(self, exc_type, exc_value, enriched_tb):
        if self._rate_limiter.please():
            sentry_sdk.capture_exception((exc_type, exc_value, enriched_tb))